#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <malloc.h>
#include <stdlib.h>    /* for qsort */
#include <stdio.h>

#include "gp2x_dir.h"

#define INITIAL_SIZE 16
#define INCREASE_SIZE 16

///////////////////////////////////

static void 
fill_entry( dir_list* listing, dir_ent* entry, const char* filename )
{
	struct stat st;
	char* path;
	
	path = ( char* ) malloc( strlen( listing->path ) + strlen( filename ) + 2 );

	strcpy( path, listing->path );
	strcat( path, "/" );
	strcat( path, filename );

	if ( stat( path, &st ) < 0 )
	{
		entry->type  = 0;
		entry->size  = 0;
		entry->perms = 0;
	}

	else
	{
		entry->perms = st.st_mode & ( S_IRWXU | S_IRWXG | S_IRWXO );

		if ( S_ISREG( st.st_mode ) )
			entry->type = '-';

		else if ( S_ISDIR( st.st_mode ) )
			entry->type = 'd';

		else if ( S_ISCHR( st.st_mode ) )
			entry->type = 'c';

		else if ( S_ISBLK( st.st_mode ) )
			entry->type = 'b';

		else if ( S_ISFIFO( st.st_mode ) )
			entry->type = 'f';

		else if ( S_ISLNK( st.st_mode ) )
			entry->type = '@';

		else if ( S_ISSOCK( st.st_mode ) )
			entry->type = 's';

		else
			entry->type = '?';

		entry->size  = st.st_size;
		entry->links = st.st_nlink;
	}

	free( path );
			
	entry->filename = strdup( filename );
}

///////////////////////////////////

static void 
dirlist_append( dir_list* listing, const char* filename )
{
    if ( !strcmp( filename, "." ) )
        return;    /* skip it */

	if ( listing->elem_count >= listing->elem_size )
	{
		dir_ent* new_elem;
		
		listing->elem_size += INCREASE_SIZE;
		new_elem = ( dir_ent * ) malloc( sizeof( dir_ent ) * listing->elem_size );

		memcpy( new_elem, listing->elements, sizeof( dir_ent ) * listing->elem_count );
		free( listing->elements );

		listing->elements = new_elem;
	}

	fill_entry( listing, listing->elements + listing->elem_count, filename );
	listing->elem_count++;
}

///////////////////////////////////

void 
dirlist_destroy( dir_list* listing )
{
	int i;
	
	for ( i = 0; i < listing->elem_count; i++ )
		free( listing->elements[i].filename );

	free( listing->elements );
	free( listing->path );
	free( listing );		
}

///////////////////////////////////

int 
dir_ent_comp( const void* d1, const void* d2 )
{
    dir_ent* de1;
    dir_ent* de2;

    de1 = ( struct dir_ent* )d1;
    de2 = ( struct dir_ent* )d2;

    if ( de1->filename && de2->filename )
        return strcmp( de1->filename, de2->filename );

    return 0;
}

///////////////////////////////////

static dir_list*
dirlist_sort( dir_list* listing )
{
    qsort( listing->elements, listing->elem_count, sizeof( dir_ent ), dir_ent_comp );

    return listing;
}

///////////////////////////////////

dir_list*
dirlist_load( const char* path )
{
	dir_list* listing;
	DIR* dir;
	struct dirent* dirent;
	
	dir = opendir( path );

	if ( dir == NULL )
		return NULL;
	
	listing             = ( dir_list * ) malloc( sizeof( dir_list ) );
	listing->path       = strdup( path );
	listing->elem_count = 0;
	listing->elem_size  = INITIAL_SIZE;
	listing->elements   = malloc( sizeof( dir_ent ) * INITIAL_SIZE );

	while ( ( dirent = readdir( dir ) ) != NULL )
		dirlist_append( listing, dirent->d_name );

	closedir( dir );

    listing = dirlist_sort( listing );

	return listing;
}

///////////////////////////////////

char* 
dirlist_get_path( dir_list *listing, int n )
{
	char *path;
	
	path = malloc( strlen( listing->path ) + strlen( listing->elements[n].filename ) + 2 );
	strcpy( path, listing->path );

	if ( strcmp( path, "/" ) != 0 )
		strcat( path, "/" );

	strcat( path, listing->elements[n].filename );

	return path;
}
