/***************************************************************************
 This is the gp2x xmame display driver.
***************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <SDL.h>
#include <SDL_ttf.h>
#include "keycodes.h"
#include "video-drivers/SDL-keytable.h"
#include "sysdep/sysdep_display_priv.h"

#include "gp2x_vkeyb.h"
#include "gp2x_opt.h"
#include "gp2x_sdl.h"

#ifdef __GP2X_HOST__
#include <SDL_rotozoom.h>
#else
#include "SDL_gp2x.h"
#endif

static SDL_Surface* video_surface     = NULL;
static SDL_Surface* offscreen_surface = NULL;
static int startx, starty;
static blit_func_p blit_func;
static int scaled_height, scaled_width;
static int first_update;
static int sdl_input_grabbed = SDL_GRAB_OFF;
static double x_scale;
static double y_scale;

/* options, these get initialised by the rc-code */
static int sdl_grab_input;
static int sdl_show_cursor;
static int sdl_always_use_mouse;
static int doublebuf;

static int sdl_mapkey( struct rc_option *option, const char *arg, int priority );

struct rc_option sysdep_display_opts[] = {
  /* name, shortname, type, dest, deflt, min, max, func, help */
  { NULL, NULL, rc_link, aspect_opts, NULL, 0, 0, NULL, NULL },
  { "SDL Related", NULL, rc_seperator, NULL, NULL, 0, 0, NULL, NULL },
  { "doublebuf", NULL, rc_bool, &doublebuf, "1", 0, 0, NULL,
    "Use double buffering to reduce flicker/tearing" },
  { "grabinput", "gi", rc_bool, &sdl_grab_input, "0", 0, 0, NULL, "Select input grabbing ( left-ctrl + delete )" },
  { "alwaysusemouse", "aum", rc_bool, &sdl_always_use_mouse, "0", 0, 0, NULL, "Always use mouse movements as input, even when not grabbed and not fullscreen ( default disabled )" },
  { "cursor", "cu", rc_bool, &sdl_show_cursor, "1", 0, 0, NULL, "Show/don't show the cursor" },
  { "sdlmapkey", "sdlmk", rc_use_function, NULL, NULL, 0, 0, sdl_mapkey,
    "Set a specific key mapping, see xmamerc.dist" },
  { NULL, NULL, rc_link, mode_opts, NULL, 0, 0, NULL, NULL },
  { NULL, NULL, rc_end, NULL, NULL, 0, 0, NULL, NULL }
};

///////////////////////////////////

int 
sysdep_display_init( void )
{
  memset( sysdep_display_properties.mode_info, 0, SYSDEP_DISPLAY_VIDEO_MODES * sizeof( int ) );
  sysdep_display_properties.mode_info[0] = SYSDEP_DISPLAY_WINDOWED | SYSDEP_DISPLAY_FULLSCREEN | SYSDEP_DISPLAY_EFFECTS;
  memset( sysdep_display_properties.mode_name, 0, SYSDEP_DISPLAY_VIDEO_MODES * sizeof( const char * ) );
  sysdep_display_properties.mode_name[0] = "SDL";

  if ( SDL_Init( SDL_INIT_VIDEO ) < 0 ) 
  {
    fprintf ( stderr, "SDL: Error: %s\n", SDL_GetError() );
    return 1;
  } 
  return 0;
}

///////////////////////////////////

void 
sysdep_display_exit( void )
{
  SDL_Quit();
}

///////////////////////////////////

static int 
SDL_calc_depth( const SDL_PixelFormat* pixel_format )
{
  return ( ( 8 - pixel_format->Rloss ) + ( 8 - pixel_format->Gloss ) + ( 8 - pixel_format->Bloss ) );
}

///////////////////////////////////

int 
sysdep_display_driver_open( int reopen )
{
  int i,j;
  int video_flags, best_window = 0;
  int best_bpp = 0, best_width = 0, best_height = 0; 
  TTF_Font* font = NULL;
  
  if ( reopen )
    sysdep_display_effect_close();
  
  video_flags = SDL_SWSURFACE;
  // video_flags = SDL_HWSURFACE | SDL_DOUBLEBUF;

  /* Find a suitable mode, also determine the max size of the display */
  scaled_height = sysdep_display_params.yarbsize ? sysdep_display_params.yarbsize : sysdep_display_params.height * sysdep_display_params.heightscale;
  scaled_width = ( ( sysdep_display_params.width + 3 ) &~ 3 ) * sysdep_display_params.widthscale;

  /* On the gp2x there is one setup, just use it rather than probing ( which also seems to crash ) */

  sysdep_display_properties.max_width  = 320;
  sysdep_display_properties.max_height = 240;

#ifndef __GP2X_HOST__

  if ( scaled_width > 320 || scaled_height > 240 )
  {
      sysdep_display_properties.max_width  = scaled_width;
      sysdep_display_properties.max_height = scaled_height;
  }

#endif

  best_width    = sysdep_display_properties.max_width;
  best_height   = sysdep_display_properties.max_height;;

  best_window   = 0;        /* bsturk: TODO: Figure out what this does */

  fprintf( stderr, "SDL: Using a mode with a resolution of: %dx%dx%d\n", best_width, best_height, 16 );

  /* Set video mode */
  if ( !video_surface )
  {
    if ( !( video_surface = SDL_SetVideoMode( best_width, best_height, 16, video_flags ) ) )
    {
      fprintf ( stderr, "SDL: Error: Setting video mode failed\n" );
      return 1;
    }

    if ( !best_window )
    {
      mode_set_aspect_ratio( ( double )best_width / best_height );
      /* mode_set_aspect_ratio may have changed yarbsize */
      scaled_height = sysdep_display_params.yarbsize ? sysdep_display_params.yarbsize: sysdep_display_params.height * sysdep_display_params.heightscale;
    }
  }

#ifdef __GP2X_HOST__
  if ( scaled_width > 320 || scaled_height > 240 )
  {
      fprintf( stderr, "gp2x: Creating an offscreen surface: %dx%dx%d\n", scaled_width, scaled_height, 32 );
      offscreen_surface = SDL_CreateRGBSurface( SDL_SWSURFACE, scaled_width, scaled_height, 32, 
              
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
		0x000000FF, 0x0000FF00, 0x00FF0000, 0 );
#else
		0xFF000000, 0x00FF0000, 0x0000FF00, 0 );
#endif
  }
#endif

  /* fill the sysdep_display_properties struct */
  memset( &sysdep_display_properties.palette_info, 0, sizeof( struct sysdep_palette_info ) );

  if ( offscreen_surface )
  {
      sysdep_display_properties.palette_info.red_mask   = offscreen_surface->format->Rmask;
      sysdep_display_properties.palette_info.green_mask = offscreen_surface->format->Gmask;
      sysdep_display_properties.palette_info.blue_mask  = offscreen_surface->format->Bmask;
      sysdep_display_properties.palette_info.depth      = SDL_calc_depth( offscreen_surface->format );
      sysdep_display_properties.palette_info.bpp        = offscreen_surface->format->BytesPerPixel * 8;
      sysdep_display_properties.vector_renderer         = NULL;
  }

  else
  {
      sysdep_display_properties.palette_info.red_mask   = video_surface->format->Rmask;
      sysdep_display_properties.palette_info.green_mask = video_surface->format->Gmask;
      sysdep_display_properties.palette_info.blue_mask  = video_surface->format->Bmask;
      sysdep_display_properties.palette_info.depth      = SDL_calc_depth( video_surface->format );
      sysdep_display_properties.palette_info.bpp        = video_surface->format->BytesPerPixel * 8;
      sysdep_display_properties.vector_renderer         = NULL;
  }

  if ( video_surface->flags & SDL_HWSURFACE )
    sysdep_display_properties.mode_info[0] |=  SYSDEP_DISPLAY_DIRECT_FB;
  else
    sysdep_display_properties.mode_info[0] &= ~SYSDEP_DISPLAY_DIRECT_FB;
  
  startx = ( video_surface->w - scaled_width ) / 2;
  starty = ( video_surface->h - scaled_height ) / 2;

  /* calculate start of screen */
  if ( scaled_width > 320 || scaled_height > 240 )
  {
#ifdef __GP2X_HOST__
      startx = 0;
      starty = 0;

      x_scale = (double)320 / (double)scaled_width;
      y_scale = (double)240 / (double)scaled_height;
#endif
  }

  if ( video_surface->flags & SDL_HWSURFACE )
    startx &= ~3;

  fprintf( stderr, "gp2x: scaled_width %d scaled_height %d\n", scaled_width, scaled_height );
  fprintf( stderr, "gp2x: widthscale %d heightscale %d yarbsize %d\n", sysdep_display_params.widthscale, sysdep_display_params.heightscale, sysdep_display_params.yarbsize );
  fprintf( stderr, "gp2x: system width %d system height %d\n", sysdep_display_params.width, sysdep_display_params.height );
  fprintf( stderr, "gp2x: startx %d starty %d\n", startx, starty );
  fflush( stderr );

  /* clear the unused area of the screen */
  for ( i = 0; i < 2; i++ )
  {
    unsigned char* video_mem;
    SDL_Surface* surface;

    if ( offscreen_surface )
        surface = offscreen_surface;
    else
        surface = video_surface;
    
    video_mem = surface->pixels;

    /* top */
    memset( video_mem, 0, starty * surface->pitch );

    /* left and right */
    for ( j = starty; j < ( scaled_height + starty ); j++ )
    {
      /* left */
      memset( video_mem + j * surface->pitch, 0,
        startx * surface->format->BytesPerPixel );

      /* right */
      memset( video_mem + j * surface->pitch +
        ( startx + scaled_width ) * surface->format->BytesPerPixel,
        0, ( surface->w - ( startx + scaled_width ) ) *
        surface->format->BytesPerPixel );
    }
    /* bottom */
    memset( video_mem + ( starty + scaled_height ) * surface->pitch, 0, 
            ( surface->h - ( starty + scaled_height ) ) * surface->pitch );
    
    SDL_UnlockSurface( surface );
    
    if ( surface->flags & SDL_DOUBLEBUF )
      SDL_Flip( surface );
    else
      break;
  }

  /* Setup input */
  if ( !reopen )
  {
    SDL_EventState( SDL_KEYUP, SDL_ENABLE );
    SDL_EventState( SDL_KEYDOWN, SDL_ENABLE );
    SDL_EnableUNICODE( 1 );
    if ( sdl_grab_input )
      sdl_input_grabbed = SDL_WM_GrabInput( SDL_GRAB_ON );
  }
  
  /* Hide/Show mouse cursor? */
  if ( ( sdl_input_grabbed == SDL_GRAB_ON ) || !sdl_show_cursor ||
      sysdep_display_params.fullscreen )
    SDL_ShowCursor( 0 );
  else
    SDL_ShowCursor( 1 );

  /* let sysdep_display_update know that it's the first call after an (re)open */
  first_update = 1;

  if ( gp2x_cur_opts.vkeyb )
  {
      font = gp2x_load_ttf_font( gp2x_cur_opts.vkeyb->font_file, gp2x_cur_opts.vkeyb->font_size );

      gp2x_cur_opts.vkeyb->font = font;
  }

  /* get a blit function */
  return !( blit_func = sysdep_display_effect_open() );
}

/*
 *  keyboard remapping routine
 *  invoiced in startup code
 *  returns 0-> success 1-> invalid from or to
 */

static int 
sdl_mapkey( struct rc_option *option, const char *arg, int priority )
{
   unsigned int from, to;
   /* ultrix sscanf() requires explicit leading of 0x for hex numbers */
   if ( sscanf( arg, "0x%x,0x%x", &from, &to ) == 2 )
   {
      /* perform tests */
      /* fprintf( stderr,"trying to map %x to %x\n", from, to ); */
      if ( from >= SDLK_FIRST && from < SDLK_LAST && to >= 0 && to <= 127 )
      {
         klookup[from] = to;
         return 0;
      }
      fprintf( stderr,"Invalid keymapping %s. Ignoring...\n", arg );
   }
   return 1;
}

///////////////////////////////////

static void
gp2x_dummy_blit(mame_bitmap *bitmap,
  rectangle *vis_in_dest_out, rectangle *dirty_area,
  struct sysdep_palette_struct *palette, unsigned char *dest, int dest_width)
{
}

///////////////////////////////////

const char *
sysdep_display_update( mame_bitmap *bitmap,
                       rectangle *vis_in_dest_out, rectangle *dirty_area,
                       struct sysdep_palette_struct *palette, int keyb_leds, int flags )
{
    unsigned char* video_mem;
    SDL_Surface* surface = NULL;
    SDL_Rect drect;

    /* do we need todo a full update? */
    if ( first_update )
    {
        *dirty_area = *vis_in_dest_out;
        first_update = 0;
    }  

    if ( offscreen_surface )
        surface = offscreen_surface;
    else
        surface = video_surface;

    if ( !( video_surface->flags & SDL_HWSURFACE ) )
        SDL_LockSurface( surface );

    video_mem  = surface->pixels;
    video_mem += startx * surface->format->BytesPerPixel;
    video_mem += starty * surface->pitch;

    blit_func( bitmap, vis_in_dest_out, dirty_area, palette, video_mem, surface->pitch / surface->format->BytesPerPixel );

    if ( !( video_surface->flags & SDL_HWSURFACE ) )
        SDL_UnlockSurface( surface );

    if ( gp2x_display_vkeyb && gp2x_cur_opts.vkeyb )
        gp2x_vkeyb_draw( surface, gp2x_cur_opts.vkeyb );

    if ( gp2x_clean_up_keyboard )
    {
        /* turned it off, clear screen */
        gp2x_vkeyb_clear_display( surface );

        gp2x_clean_up_keyboard--;
    }

#ifdef __GP2X_HOST__
    if ( offscreen_surface )
    {
        SDL_Rect sr, dr;

        dr.x = startx + vis_in_dest_out->min_x;
        dr.y = starty + vis_in_dest_out->min_y;
        dr.w = vis_in_dest_out->max_x - vis_in_dest_out->min_x;
        dr.h = vis_in_dest_out->max_y - vis_in_dest_out->min_y;

        sr.x = 0;
        sr.y = 0;
        sr.w = video_surface->w;
        sr.h = video_surface->h;

        SDL_Surface* stretched_surface = zoomSurface( offscreen_surface, x_scale, y_scale, 1 );

        SDL_BlitSurface( stretched_surface, &sr, video_surface, &dr );
    }
#endif

    if ( video_surface->flags & SDL_DOUBLEBUF )
      SDL_Flip( video_surface );

    if ( !( video_surface->flags & SDL_HWSURFACE ) )
        SDL_UpdateRect( video_surface, 0, 0, 0, 0 );

    return NULL;
}

///////////////////////////////////

void 
sysdep_display_driver_clear_buffer( void )
{
  int i, line;
  unsigned char *video_mem;

  for ( i = 0; i < 2; i++ )
  {
    SDL_LockSurface( video_surface );

    video_mem = video_surface->pixels;
    video_mem += startx * video_surface->format->BytesPerPixel;
    video_mem += starty * video_surface->pitch;

    for ( line=0; line<scaled_height; line++ )
      memset( video_mem + line*video_surface->pitch, 0,
        scaled_width*video_surface->format->BytesPerPixel );

    SDL_UnlockSurface( video_surface );

    if ( video_surface->flags & SDL_DOUBLEBUF )
      SDL_Flip( video_surface );
    else
      break;
  }
}

/* shut up the display */
void sysdep_display_close( void )
{
   sysdep_display_effect_close();
}

void sysdep_display_update_mouse( void )
{
   int i;
   int x,y;
   Uint8 buttons;

   if( sdl_always_use_mouse || ( sdl_input_grabbed == SDL_GRAB_ON ) ||
      sysdep_display_params.fullscreen )
   {
     buttons = SDL_GetRelativeMouseState( &x, &y );
     sysdep_display_mouse_data[0].deltas[0] = x;
     sysdep_display_mouse_data[0].deltas[1] = y;
     for( i=0;i<SYSDEP_DISPLAY_MOUSE_BUTTONS;i++ ) {
        sysdep_display_mouse_data[0].buttons[i] = buttons & ( 0x01 << i );
     }
   }
   else
   {
     sysdep_display_mouse_data[0].deltas[0] = 0;
     sysdep_display_mouse_data[0].deltas[1] = 0;
     for( i=0;i<SYSDEP_DISPLAY_MOUSE_BUTTONS;i++ ) {
        sysdep_display_mouse_data[0].buttons[i] = 0;
     }
   }
}

///////////////////////////////////

int 
sysdep_display_driver_update_keyboard() 
{
   struct sysdep_display_keyboard_event kevent;
   SDL_Event event;
   int retval = 0;

   if ( video_surface ) 
   {
#ifndef __GP2X_HOST__
       if ( !gp2x_display_vkeyb )
       {
#endif
          while( SDL_PollEvent( &event ) ) 
          {
             kevent.press = 0;
             
             switch ( event.type )
             {
                case SDL_KEYDOWN:
                   kevent.press = 1;
                case SDL_KEYUP:
                   kevent.scancode = klookup[event.key.keysym.sym];
                   kevent.unicode = event.key.keysym.unicode;
                   sysdep_display_params.keyboard_handler( &kevent );
                   break;
                case SDL_QUIT:
                   retval |= SYSDEP_DISPLAY_QUIT_REQUESTED;
                   break;
                default:
                   break;
             }
          }
#ifndef __GP2X_HOST__
       }
#endif
   }

   return retval;
}
